--[[	PhantomPlates Options Module
	by SDPhantom
	https://www.wowinterface.com/forums/member.php?u=34145
	https://www.curseforge.com/members/sdphantomgamer/projects	]]
--------------------------------------------------------------------------

--------------------------
--[[	Namespace	]]
--------------------------
local Name,AddOn=...;
local Title=select(2,GetAddOnInfo(Name)):gsub("%s*v?[%d%.]+$","");
local Version=GetAddOnMetadata(Name,"Version");
local Author=GetAddOnMetadata(Name,"Author");

local OptionsPanel=CreateFrame("Frame");
AddOn.Options_OptionsPanel=OptionsPanel;
OptionsPanel:Hide();

--------------------------
--[[	Options Tables	]]
--------------------------
local Defaults,Options={},nil;
AddOn.Options=Defaults;--	Modules can write to defaults while loading

--------------------------
--[[	Local Variables	]]
--------------------------
local OptionsCache={};

local OptionsLayoutSectionAnchors={
	{LEFT={"LEFT",8,0},RIGHT={"CENTER",-2,0}};
	{LEFT={"CENTER",2,0},RIGHT={"RIGHT",-8,0}};
};
OptionsLayoutSectionObjects={};
for i=1,#OptionsLayoutSectionAnchors do OptionsLayoutSectionObjects[i]={}; end

local OptionsPanelControls=setmetatable({},{__index=function(t,k)
	local proxy={Key=k};
	t[k]=proxy; return proxy;
end});

local OptionsPanelControlDataMap={};

----------------------------------
--[[	Options Panel Layout	]]
----------------------------------
do	local title=OptionsPanel:CreateFontString(nil,"OVERLAY","GameFontNormalLarge");
	title:SetPoint("TOP",0,-12);
	title:SetText(Title);

	local author=OptionsPanel:CreateFontString(nil,"OVERLAY","GameFontNormalSmall");
	author:SetPoint("TOP",title,"BOTTOM");
	author:SetTextColor(1,0.5,0.25);
	author:SetFormattedText(AddOn.Localization.Options_ByAuthor_Format,Author);

	local ver=OptionsPanel:CreateFontString(nil,"OVERLAY","GameFontNormalSmall");
	ver:SetPoint("TOPLEFT",title,"TOPRIGHT",4,0);
	ver:SetTextColor(0.5,0.5,0.5);
	ver:SetText("v"..Version);
end

------------------------------------------
--[[	Options Panel Event Handler	]]
------------------------------------------
OptionsPanel:RegisterEvent("ADDON_LOADED");
OptionsPanel:SetScript("OnEvent",function(self,event,...)
	if event=="ADDON_LOADED" and (...)==Name then
--		Load Options
		local rawoptions=PhantomPlates_Options;
		if not rawoptions then rawoptions={}; PhantomPlates_Options=rawoptions; end

--		Trim/Fill Options
--		for key,val in pairs(rawoptions) do if Defaults[key]==nil then rawoptions[key]=nil; end end
		for key,val in pairs(Defaults) do if rawoptions[key]==nil then rawoptions[key]=val; end end

--		Attach Metatable
		Options=setmetatable({},{
			__index=rawoptions;
			__newindex=function(t,k,v)
				if rawoptions[k]~=v then
					rawoptions[k]=v;
					local data=OptionsPanelControls[k];
					if data.OnValueChanged then
						data.OnValueChanged(data.Control,k,v);
					end
				end
			end;
		});
		AddOn.Options=Options;		

--		Send updates to modules
		for key,data in pairs(OptionsPanelControls) do
			if data.OnValueChanged then
				data.OnValueChanged(data.Control,key,rawoptions[key]);
			end
		end

--		Don't need this event anymore
		self:UnregisterEvent(event);
	end
end);

----------------------------------
--[[	Options Panel Callbacks	]]
----------------------------------
local function OptionsPanel_Okay()
	table.wipe(OptionsCache);--	Options already set by preview, just cleanup
end

local function OptionsPanel_Cancel()
	for key,val in pairs(OptionsCache) do
		if Options[key]~=val then Options[key]=val; end--	Restore previous setting
	end
	table.wipe(OptionsCache);--	Cleanup
end

local function OptionsPanel_Refresh()
	table.wipe(OptionsCache);--	Sanity check

	for key,data in pairs(OptionsPanelControls) do
		if data.Control then
			if data.Refresh then data.Refresh(data.Control); end
			if data.SetValue then data.SetValue(data.Control,Options[key]); end
		end
	end
end

local function OptionsPanel_Default()
	for key,data in pairs(OptionsPanelControls) do
		local val=Defaults[key]; if Options[key]~=val then
			Options[key]=val;--	Restore default
			if data.Control and data.SetValue then data.SetValue(data.Control,val); end
		end
	end
end

------------------------------------------
--[[	Options Panel Registration	]]
------------------------------------------
if Settings and (Settings.RegisterCanvasLayoutCategory and Settings.RegisterAddOnCategory) then
	OptionsPanel.OnCommit=OptionsPanel_Okay;
	OptionsPanel.OnDefault=OptionsPanel_Default;
	OptionsPanel.OnRefresh=OptionsPanel_Refresh;

	local category=Settings.RegisterCanvasLayoutCategory(OptionsPanel,Title);
	category.ID=Title;
	Settings.RegisterAddOnCategory(category);
else
	OptionsPanel.name=Title;
	OptionsPanel.okay=OptionsPanel_Okay;
	OptionsPanel.cancel=OptionsPanel_Cancel;
	OptionsPanel.default=OptionsPanel_Default;
	OptionsPanel.refresh=OptionsPanel_Refresh;
	InterfaceOptions_AddCategory(OptionsPanel);
end

--------------------------
--[[	Options API	]]
--------------------------
--	Options Panel Layout
function AddOn.Options_CreateSubPanel(sectionindex)
	sectionindex=sectionindex and Clamp(sectionindex,1,#OptionsLayoutSectionObjects);
	local sectionobj=sectionindex and OptionsLayoutSectionObjects[sectionindex];
	if not (sectionindex and sectionobj) then
		local lowheight; for index,objlist in ipairs(OptionsLayoutSectionObjects) do
			local height=0; for _,obj in ipairs(objlist) do
				height=height+obj:GetHeight();
			end
			if not lowheight or height<lowheight then
				sectionindex,sectionobj,lowheight=index,objlist,height;
			end
		end
	end

	local subpanel=CreateFrame("Frame",nil,OptionsPanel,DoesTemplateExist("BackdropTemplate") and "BackdropTemplate" or nil);
	subpanel:SetBackdrop(BACKDROP_TOOLTIP_0_16_5555 or {
		edgeFile="Interface\\Tooltips\\UI-Tooltip-Border";
		edgeSize=16;
		tileEdge=true;
		insets={left=5,right=5,top=5,bottom=5};
	});
	subpanel:SetBackdropBorderColor(LIGHTGRAY_FONT_COLOR:GetRGB());

--	Top Anchor
	local len=#sectionobj;
	if len>0 then subpanel:SetPoint("TOP",sectionobj[len],"BOTTOM",0,-4);
	else subpanel:SetPoint("TOP",0,-48); end

--	Side Anchors
	for point,anchor in pairs(OptionsLayoutSectionAnchors[sectionindex]) do
		if point==anchor[1] then subpanel:SetPoint(unpack(anchor));
		else subpanel:SetPoint(point,OptionsPanel,unpack(anchor)); end
	end

	table.insert(sectionobj,subpanel);
	return subpanel;
end

--	Options Panel Controls
local function OptionsControlObject_OnUpdateScript(self)
	local data=OptionsPanelControlDataMap[self];
	if data.GetValue then
		local key,val=data.Key,data.GetValue(self);
		if Options[key]~=val then Options[key]=val; end
	end
end

local function OptionsControlObject_OnEnable(self)
	local data=OptionsPanelControlDataMap[self];

	if data.Children then
		for _,child in ipairs(data.Children) do child:Enable(); end
	end

	if data.TextObjects then
		for text,stateinfo in pairs(data.TextObjects) do
			if stateinfo.EnabledObj then
				text:SetFontObject(stateinfo.EnabledObj);
			else text:SetTextColor(stateinfo.EnabledR,stateinfo.EnabledG,stateinfo.EnabledB); end
		end
	end

	if data.OnEnable then data.OnEnable(self); end
end

local function OptionsControlObject_OnDisable(self)
	local data=OptionsPanelControlDataMap[self];

	if data.Children then
		for _,child in ipairs(data.Children) do child:Disable(); end
	end

	if data.TextObjects then
		for text,stateinfo in pairs(data.TextObjects) do
			if stateinfo.DisabledObj then
				text:SetFontObject(stateinfo.DisabledObj);
			else text:SetTextColor(stateinfo.DisabledR,stateinfo.DisabledG,stateinfo.DisabledB); end
		end
	end

	if data.OnDisable then data.OnDisable(self); end
end

function AddOn.Options_RegisterControl(key,control,refreshfunc,setfunc,getfunc,updatescript)
	local data=OptionsPanelControls[key];--	Metatable creates basic proxy as needed
	OptionsPanelControlDataMap[control]=data;

	data.Control=control;
	data.Refresh=refreshfunc;
	data.SetValue=setfunc;
	data.GetValue=getfunc;

	if updatescript then control:HookScript(updatescript,OptionsControlObject_OnUpdateScript); end
	if control.SetEnabled then
		control:HookScript("OnEnable",OptionsControlObject_OnEnable);
		control:HookScript("OnDisable",OptionsControlObject_OnDisable);
	end
end

function AddOn.Options_GetControlFromOptionKey(key)
	local data=rawget(OptionsPanelControls,key);
	return data and data.Control;
end

function AddOn.Options_GetOptionKeyFromControl(control)
	local data=OptionsPanelControlDataMap[control];
	return data and data.Key;
end

function AddOn.Options_AddChildToControl(key,control)
	local data=OptionsPanelControls[key];--	Metatable creates basic proxy as needed
	if data.Children then
		tInsertUnique(data.Children,control);
	else data.Children={control}; end
end

function AddOn.Options_AddTextToControl(key,text,...)
	local eobj,er,eg,eb,dobj,er,eg,eb; do
		er,eg,eb,dr,dg,db=...;

		if type(er)=="table" then
			dr,dg,db=eg,eb,dr;--	Arg shift
			if type(er[0])=="userdata" then eobj,er,eg,eb=er,nil,nil,nil;
			elseif er.r and er.g and er.b then eobj,er,eg,eb=nil,er.r,er.g,er.b;
			elseif #er>=3 then eobj,er,eg,eb=nil,unpack(er,1,3);
			else error("Invalid enabled color: Expected (R,G,B), color table, or font object"); end
		end

		if type(dr)=="table" then
			if type(dr[0])=="userdata" then dobj,dr,dg,db=dr,nil,nil,nil;
			elseif dr.r and dr.g and dr.b then dobj,dr,dg,db=nil,dr.r,dr.g,dr.b;
			elseif #dr>=3 then dobj,dr,dg,db=nil,unpack(dr,1,3);
			else error("Invalid enabled color: Expected (R,G,B), color table, or font object"); end
		end
	end

	local stateinfo={
		EnabledObj=eobj;
		EnabledR=er,EnabledG=eg,EnabledB=eb;

		DisabledObj=dobj;
		DisabledR=dr,DisabledG=dg,DisabledB=db;
	};

	local data=OptionsPanelControls[key];--	Metatable creates basic proxy as needed
	if data.TextObjects then
		data.TextObjects[text]=stateinfo;
	else data.TextObjects={[text]=stateinfo}; end
end

function AddOn.Options_SetControlEnabled(key,enabled)
	local data=rawget(OptionsPanelControls,key);
	if data and data.Control then
		if data.Control.SetEnabled then
			data.Control:SetEnabled(enabled);
		else
			(	enabled
			and	OptionsControlObject_OnEnable
			or	OptionsControlObject_OnDisable
			)(data.Control);
		end
	end
end

local function HookOptionCallback(key,event,func)
	local data=OptionsPanelControls[key];--	Metatable creates basic proxy as needed
	if not data[event] then data[event]=AddOn.CreateDispatcher(); end
	return data[event]:AddCallback(func);
end

function AddOn.Options_HookOption(key,func) return HookOptionCallback(key,"OnValueChanged",func); end
function AddOn.Options_HookOnEnable(key,func) return HookOptionCallback(key,"OnEnable",func); end
function AddOn.Options_HookOnDisable(key,func) return HookOptionCallback(key,"OnDisable",func); end
